#ifndef TSofMwpcPHYSICS_H
#define TSofMwpcPHYSICS_H
/*****************************************************************************
 * Copyright (C) 2009-2020   this file is part of the NPTool Project       *
 *                                                                           *
 * For the licensing terms see $NPTOOL/Licence/NPTool_Licence                *
 * For the list of contributors see $NPTOOL/Licence/Contributors             *
 *****************************************************************************/

/*****************************************************************************
 * Original Author: Pierre Morfouace  contact address: pierre.morfouace2@cea.fr                        *
 *                                                                           *
 * Creation Date  : November 2020                                           *
 * Last update    :                                                          *
 *---------------------------------------------------------------------------*
 * Decription:                                                               *
 *  This class hold SofMwpc Treated data                                *
 *                                                                           *
 *---------------------------------------------------------------------------*
 * Comment:                                                                  *
 *                                                                           *   
 *                                                                           *
 *****************************************************************************/

// C++ headers 
#include <vector>
#include <map>
#include <string>
using namespace std;

// ROOT headers
#include "TObject.h"
#include "TH1.h"
#include "TVector3.h"
#include "TSpline.h"
// NPTool headers
#include "TSofMwpcData.h"
#include "NPCalibrationManager.h"
#include "NPVDetector.h"
#include "NPInputParser.h"



class TSofMwpcPhysics : public TObject, public NPL::VDetector {
  //////////////////////////////////////////////////////////////
  // constructor and destructor
  public:
    TSofMwpcPhysics();
    ~TSofMwpcPhysics() {};


  //////////////////////////////////////////////////////////////
  // Inherited from TObject and overriden to avoid warnings
  public: 
    void Clear();   
    void Clear(const Option_t*) {};


  //////////////////////////////////////////////////////////////
  // data obtained after BuildPhysicalEvent() and stored in
  // output ROOT file
  public:
    vector<int>      DetectorNbr;
    vector<double>   PositionX1;
    vector<double>   PositionX2;
    vector<double>   PositionY;
  
  private:
    vector<double>   DetPosX;//!
    vector<double>   DetPosY;//!
    vector<double>   DetPosZ;//!
    vector<vector<double>> Buffer_X1_Q;//!
    vector<vector<double>> Buffer_X2_Q;//!
    vector<vector<double>> Buffer_Y_Q;//!
  
    /// A usefull method to bundle all operation to add a detector
  void AddDetector(TVector3 Pos); 
  void AddDetector(double R, double Theta, double Phi); 
  
  //////////////////////////////////////////////////////////////
  // methods inherited from the VDetector ABC class
  public:
    // read stream from ConfigFile to pick-up detector parameters
    void ReadConfiguration(NPL::InputParser);

    // add parameters to the CalibrationManger
    void AddParameterToCalibrationManager();

    // method called event by event, aiming at extracting the 
    // physical information from detector
    void BuildPhysicalEvent();

    // same as BuildPhysicalEvent() method but with a simpler
    // treatment
    void BuildSimplePhysicalEvent();

    // same as above but for online analysis
    void BuildOnlinePhysicalEvent()  {BuildPhysicalEvent();};

    // activate raw data object and branches from input TChain
    // in this method mother branches (Detector) AND daughter leaves 
    // (fDetector_parameter) have to be activated
    void InitializeRootInputRaw();

    // activate physics data object and branches from input TChain
    // in this method mother branches (Detector) AND daughter leaves 
    // (fDetector_parameter) have to be activated
    void InitializeRootInputPhysics();

    // create branches of output ROOT file
    void InitializeRootOutput();

    // clear the raw and physical data objects event by event
    void ClearEventPhysics() {Clear();}      
    void ClearEventData()    {m_EventData->Clear();}   


  //////////////////////////////////////////////////////////////
  // specific methods to SofMwpc array
  public:
    // remove bad channels, calibrate the data and apply thresholds
    void PreTreat();

    // clear the pre-treated object
    void ClearPreTreatedData()   {m_PreTreatedData->Clear();}

    // read the user configuration file. If no file is found, load standard one
    void ReadAnalysisConfig();

    // give and external TSofMwpcData object to TSofMwpcPhysics. 
    // needed for online analysis for example
    void SetRawDataPointer(TSofMwpcData* rawDataPointer) {m_EventData = rawDataPointer;}

    double fwx[4]={3.125, 3.125, 3.125, 3.125};
    double fwy[4]={3.125, 5.000, 5.000, 5.000};
    double fSizeX[4]={200, 200, 200, 900};
    double fSizeY[4]={200, 200, 200, 600};
  
    vector<pair<int, int>> fPairX;
    vector<pair<int, int>> fPairY;
    vector<vector<pair<int,int>>> fClusterX;
    vector<vector<pair<int,int>>> fClusterY;
    
    double fThresholdX;
    double fThresholdY;

    double GetChargeMax(vector<pair<int,int>> p1);
    vector<pair<int,int>> FindCluster(vector<pair<int,int>>& p1);

    double GetPositionX(int det, double qmax, int padmax, double qleft, double qright);
    double GetPositionY(int det, double qmax, int padmax, double qdown, double qup);
    
  // objects are not written in the TTree
  private:
    TSofMwpcData*         m_EventData;        //!
    TSofMwpcData*         m_PreTreatedData;   //!
    TSofMwpcPhysics*      m_EventPhysics;     //!

  // getters for raw and pre-treated data object
  public:
    TSofMwpcData* GetRawData()        const {return m_EventData;}
    TSofMwpcData* GetPreTreatedData() const {return m_PreTreatedData;}

  // parameters used in the analysis
  private:
    double m_E_Threshold;     //!

  // number of detectors
  private:
    int m_NumberOfDetectors;  //!

  // Static constructor to be passed to the Detector Factory
  public:
    static NPL::VDetector* Construct();

    ClassDef(TSofMwpcPhysics,1)  // SofMwpcPhysics structure
};
#endif
